'use strict'

import bsock = require('bsock');

import * as T from './Types'; 
import * as I from './Interfaces';

//fix args with defaults like "force = true" -> "force"
function stripAfterEquals(str:string){
    return str.split("=")[0]
}

export class RPCSocket implements I.Socket{

    private socket: I.Socket
    constructor(public port:number, private server: string, private tls: boolean = false){
        Object.defineProperty(this, 'socket', {value: undefined, writable: true})
    }

    public hook(name: T.Name, args: T.Arg){
        return this.socket.hook(name, args)
    }

    public unhook(name: T.Name){
        return this.socket.unhook(name)
    }

    public on(type: "error" | "close", f: (e?: any) => void){
        return this.socket.on(type, f)
    }

    public destroy(){
        return this.socket.destroy()
    }

    public close(){
        return this.socket.close()
    }

    public async call (rpcname: T.Name, ...args: T.Any[]) : Promise<T.Any>{
        return await this.socket.call.apply(this.socket, [rpcname, ...args])
    }

    public async fire(rpcname: T.Name, ...args: T.Any[]) : Promise<T.Any>{
        return await this.socket.fire.apply(this.socket, [rpcname, ...args])
    }
    
    public async connect(){
        this.socket = await bsock.connect(this.port, this.server, this.tls)
        this.on('error', () => {})

        const info:T.ExtendedRpcInfo[] = await this.info()
        info.forEach(i => {
            let f: any
            switch (i.type) {
                case 'Call':
                    f = this.callGenerator(i.uniqueName, i.argNames)    
                    break                
                case 'Hook':
                    f = this.hookGenerator(i.uniqueName, i.argNames)                    
                    break
            }
            if(this[i.owner] == null)
                this[i.owner] = {}
            this[i.owner][i.name] = f
            this[i.owner][i.name].bind(this)
        })
    }

    public async info(){
        return await this.socket.call('info')
    }

    private callGenerator(fnName: T.Name, fnArgs:T.Arg[]): T.AsyncFunction{
        const headerArgs = fnArgs.join(",")
        const argParams = fnArgs.map(stripAfterEquals).join(",")
        return eval( '( () => async ('+headerArgs+') => { return await this.socket.call("'+fnName+'", '+argParams+')} )()' )
    }

    private hookGenerator(fnName: T.Name, fnArgs:T.Arg[]): T.HookFunction{
        const headerArgs = fnArgs.join(",")
        const argParams = fnArgs.map(stripAfterEquals).join(",")
        return eval( `( () => async (`+headerArgs+(headerArgs.length!==0?",":"")+` callback) => {
                            const r = await this.socket.call("`+fnName+`", `+argParams+`)
                            if(r.result === 'Success'){
                                this.socket.hook(r.uuid, callback)
                            }
                            return r
                        } )()` )
    }
}